@@ -2,12 +2,15 @@ import argparse
 import glob
 import logging
 import os
+import shutil
 import sys
 import textwrap
 import warnings
+from typing import Tuple
 
 from . import __version__
 from .config import (
+    COMPILED_DIR,
     get_bustools_binary_path,
     get_kallisto_binary_path,
     is_dry,
@@ -19,9 +22,11 @@ from .config import (
     set_kallisto_binary_path,
     TECHNOLOGIES,
     TEMP_DIR,
+    UnsupportedOSError,
 )
+from .compile import compile
 from .constants import INFO_FILENAME
-from .count import count, count_smartseq, count_velocity
+from .count import count, count_smartseq, count_smartseq3, count_velocity
 from .logging import logger
 from .ref import download_reference, ref, ref_kite, ref_lamanno
 from .utils import (
@@ -33,16 +38,49 @@ from .utils import (
 )
 
 
+def test_binaries() -> Tuple[bool, bool]:
+    """Test whether kallisto and bustools binaries are executable.
+
+    Internally, this function calls :func:`utils.get_kallisto_version` and
+    :func:`utils.get_bustools_version`, both of which return `None` if there is
+    something wrong with their respective binaries.
+
+    Returns:
+        A tuple of two booleans indicating kallisto and bustools binaries.
+    """
+    kallisto_ok = True
+    try:
+        kallisto_ok = get_kallisto_version() is not None
+    except Exception:
+        kallisto_ok = False
+    bustools_ok = True
+    try:
+        bustools_ok = get_bustools_version() is not None
+    except Exception:
+        bustools_ok = False
+
+    return kallisto_ok, bustools_ok
+
+
+def get_binary_info() -> str:
+    """Get information on the binaries that will be used for commands.
+
+    Returns:
+        `kallisto` and `bustools` binary versions and paths.
+    """
+    kallisto_version = '.'.join(str(i) for i in get_kallisto_version())
+    bustools_version = '.'.join(str(i) for i in get_bustools_version())
+    return (
+        f'kallisto: {kallisto_version} ({get_kallisto_binary_path()})\n'
+        f'bustools: {bustools_version} ({get_bustools_binary_path()})'
+    )
+
+
 def display_info():
     """Displays kb, kallisto and bustools version + citation information, along
     with a brief description and examples.
     """
-    kallisto_version = '.'.join(str(i) for i in get_kallisto_version())
-    bustools_version = '.'.join(str(i) for i in get_bustools_version())
-    info = '''kb_python {}
-    kallisto: {}
-    bustools: {}
-    '''.format(__version__, kallisto_version, bustools_version)
+    info = f'kb_python {__version__}\n{get_binary_info()}'
     with open(os.path.join(PACKAGE_PATH, INFO_FILENAME), 'r') as f:
         print(
             '{}\n{}'.format(
@@ -60,7 +98,7 @@ def display_technologies():
     """Displays a list of supported technologies along with whether kb provides
     a whitelist for that technology and the FASTQ argument order for kb count.
     """
-    headers = ['name', 'whitelist', 'barcode', 'umi', 'cDNA']
+    headers = ['name', 'description', 'whitelist', 'barcode', 'umi', 'cDNA']
     rows = [headers]
 
     print('List of supported single-cell technologies\n')
@@ -74,9 +112,10 @@ def display_technologies():
         chem = t.chemistry
         row = [
             t.name,
+            t.description,
             'yes' if chem.has_whitelist else '',
-            ' '.join(str(_def) for _def in chem.cell_barcode_parser)
-            if chem.has_cell_barcode else '',
+            ' '.join(str(_def) for _def in chem.barcode_parser)
+            if chem.has_barcode else '',
             ' '.join(str(_def)
                      for _def in chem.umi_parser) if chem.has_umi else '',
             ' '.join(str(_def) for _def in chem.cdna_parser),
@@ -97,11 +136,74 @@ def display_technologies():
     sys.exit(1)
 
 
-def parse_ref(parser, args, temp_dir='tmp'):
+def parse_compile(
+    parser: argparse.ArgumentParser,
+    args: argparse.Namespace,
+    temp_dir: str = 'tmp'
+):
+    """Parser for the `compile` command.
+
+    Args:
+        parser: The argument parser
+        args: Parsed command-line arguments
+    """
+    # target must be all when --view is used
+    if args.view and args.target is not None:
+        parser.error(
+            '`target` must not be be provided when `--view` is provided.'
+        )
+
+    # target must not be all when --url is provided
+    if args.url and args.target == 'all':
+        parser.error('`target` must not be `all` when `--url` is provided.')
+
+    # --view or --remove may not be specified with -o
+    if args.o and (args.view or args.remove):
+        parser.error('`-o` may not be used with `--view` or `--remove`')
+    if args.cmake_arguments and (args.view or args.remove):
+        parser.error(
+            '`--cmake-arguments` may not be used with `--view` or `--remove`'
+        )
+
+    if args.remove:
+        if args.target in ('kallisto', 'all'):
+            shutil.rmtree(
+                os.path.join(COMPILED_DIR, 'kallisto'), ignore_errors=True
+            )
+        if args.target in ('bustools', 'all'):
+            shutil.rmtree(
+                os.path.join(COMPILED_DIR, 'bustools'), ignore_errors=True
+            )
+    elif args.view:
+        print(get_binary_info())
+        sys.exit(1)
+    else:
+        if args.target not in ('kallisto', 'bustools', 'all'):
+            parser.error(
+                '`target` must be one of `kallisto`, `bustools`, `all`'
+            )
+
+        compile(
+            args.target,
+            out_dir=args.o,
+            cmake_arguments=args.cmake_arguments,
+            url=args.url,
+            ref=args.ref,
+            overwrite=args.overwrite,
+            temp_dir=temp_dir
+        )
+
+
+def parse_ref(
+    parser: argparse.ArgumentParser,
+    args: argparse.Namespace,
+    temp_dir: str = 'tmp'
+):
     """Parser for the `ref` command.
 
-    :param args: Command-line arguments dictionary, as parsed by argparse
-    :type args: dict
+    Args:
+        parser: The argument parser
+        args: Parsed command-line arguments
     """
     if args.k is not None:
         if args.k < 0 or not args.k % 2:
@@ -115,6 +217,22 @@ def parse_ref(parser, args, temp_dir='tmp'):
             'There must be the same number of FASTAs as there are GTFs.'
         )
 
+    # Parse include/exclude KEY:VALUE pairs
+    include = []
+    exclude = []
+    if args.include_attribute:
+        for kv in args.include_attribute:
+            key, value = kv.split(':')
+            if kv.count(':') != 1 or not key or not value:
+                parser.error(f'Malformed KEY:VALUE pair `{kv}`')
+            include.append({key: value})
+    if args.exclude_attribute:
+        for kv in args.exclude_attribute:
+            key, value = kv.split(':')
+            if kv.count(':') != 1 or not key or not value:
+                parser.error(f'Malformed KEY:VALUE pair `{kv}`')
+            exclude.append({key: value})
+
     if args.d is not None:
         # Options that are files.
         options = ['i', 'g', 'c1', 'c2']
@@ -140,6 +258,8 @@ def parse_ref(parser, args, temp_dir='tmp'):
             n=args.n,
             k=args.k,
             flank=args.flank,
+            include=include,
+            exclude=exclude,
             overwrite=args.overwrite,
             temp_dir=temp_dir
         )
@@ -153,6 +273,12 @@ def parse_ref(parser, args, temp_dir='tmp'):
                 )
 
         if args.workflow == 'kite':
+            if args.include_attribute or args.exclude_attribute:
+                parser.error(
+                    '`--include-attribute` or `--exclude-attribute` may not be used '
+                    f'for workflow `{args.workflow}`'
+                )
+
             ref_kite(
                 args.feature,
                 args.f1,
@@ -173,16 +299,23 @@ def parse_ref(parser, args, temp_dir='tmp'):
                 args.g,
                 n=args.n,
                 k=args.k,
+                include=include,
+                exclude=exclude,
                 overwrite=args.overwrite,
                 temp_dir=temp_dir
             )
 
 
-def parse_count(parser, args, temp_dir='tmp'):
+def parse_count(
+    parser: argparse.ArgumentParser,
+    args: argparse.Namespace,
+    temp_dir: str = 'tmp'
+):
     """Parser for the `count` command.
 
-    :param args: Command-line arguments dictionary, as parsed by argparse
-    :type args: dict
+    Args:
+        parser: The argument parser
+        args: Parsed command-line arguments
     """
     if args.report:
         logger.warning((
@@ -202,9 +335,143 @@ def parse_count(parser, args, temp_dir='tmp'):
     if args.w and args.w.lower() == 'none':
         args.w = None
 
+    if args.filter_threshold and args.filter != 'bustools':
+        parser.error(
+            'Option `--filter-threshold` may only be used with `--filter bustools`.'
+        )
+
+    if args.tcc and args.cellranger:
+        parser.error(
+            'TCC matrices can not be converted to cellranger-compatible format.'
+        )
+    if args.tcc and args.report:
+        logger.warning(
+            'Plots for TCC matrices have not yet been implemented. '
+            'The HTML report will not contain any plots.'
+        )
+    if args.tcc and args.em:
+        parser.error('`--tcc` may not be used with `--em`.')
+    if args.gene_names and not (args.loom or args.h5ad):
+        parser.error(
+            '`--gene-names` may only be used with `--h5ad` or `--loom`'
+        )
+    if args.tcc and args.gene_names:
+        parser.error('`--gene-names` may not be used with `--tcc`')
+
+    # Check if batch TSV was provided.
+    batch_path = None
+    if len(args.fastqs) == 1:
+        try:
+            with open_as_text(args.fastqs[0], 'r') as f:
+                if not f.readline().startswith('@'):
+                    batch_path = args.fastqs[0]
+        except Exception:
+            pass
+
+    if args.x.upper() in ('BULK', 'SMARTSEQ2', 'SMARTSEQ3') and (args.umi_gene
+                                                                 or args.em):
+        parser.error(
+            f'`--umi-gene` or `--em` may not be used for technology {args.x}'
+        )
+    if args.x.upper() in ('BULK', 'SMARTSEQ2'):
+        # Check unsupported options
+        unsupported = ['filter', 'w']
+        for arg in unsupported:
+            if getattr(args, arg):
+                parser.error(
+                    f'Argument `{arg}` is not supported for technology `{args.x}`.'
+                )
+
+        if not args.parity:
+            parser.error(
+                f'`--parity` must be provided for technology `{args.x}`.'
+            )
+
+        if not batch_path:
+            logger.warning(
+                f'FASTQs were provided for technology `{args.x}`. '
+                'Assuming multiplexed samples. For demultiplexed samples, provide '
+                'a batch textfile.'
+            )
+        else:
+            # If `single`, then each row must contain 2 columns. If `paired`,
+            # each row must contain 3 columns.
+            target = 2 + (args.parity == 'paired')
+            with open(batch_path, 'r') as f:
+                for i, line in enumerate(f):
+                    if line.isspace() or line.startswith('#'):
+                        continue
+                    sep = '\t' if '\t' in line else ' '
+                    columns = len(line.split(sep))
+                    if target != columns:
+                        parser.error(
+                            f'Batch file {batch_path} line {i} contains wrong '
+                            f'number of columns. Expected {target} for '
+                            f'`--parity {args.parity} but got {columns}.'
+                        )
+
+        if args.parity == 'single':
+            if args.tcc:
+                if (args.fragment_l is None) ^ (args.fragment_s is None):
+                    parser.error(
+                        'Both or neither `--fragment-l` and `--fragment-s` must be '
+                        'provided for single-end reads with TCC output.'
+                    )
+
+                if args.fragment_l is None and args.fragment_s is None:
+                    logger.warning(
+                        '`--fragment-l` and `--fragment-s` not provided. '
+                        'Assuming all transcripts have the exact same length.'
+                    )
+            elif (args.fragment_l is not None) or (args.fragment_s is not None):
+                parser.error(
+                    '`--fragment-l` and `--fragment-s` may only be used with `--tcc`.'
+                )
+
+        elif args.parity == 'paired':
+            if args.fragment_l is not None or args.fragment_s is not None:
+                parser.error(
+                    '`--fragment-l` or `--fragment-s` may not be provided for '
+                    'paired-end reads.'
+                )
+    elif args.x.upper() == 'SMARTSEQ3':
+        unsupported = [
+            'filter', 'w', 'parity', 'fragment-l', 'fragment-s', 'report',
+            'cellranger'
+        ]
+        for arg in unsupported:
+            if getattr(args, arg.replace('-', '_')):
+                parser.error(
+                    f'Argument `{arg}` is not supported for technology `{args.x}`.'
+                )
+
+            # Batch file not supported
+            if batch_path:
+                parser.error(
+                    f'Technology {args.x} does not support a batch file.'
+                )
+    else:
+        # Check unsupported options
+        unsupported = ['parity', 'fragment-l', 'fragment-s']
+        for arg in unsupported:
+            if getattr(args, arg.replace('-', '_')):
+                parser.error(
+                    f'Argument `{arg}` is not supported for technology `{args.x}`.'
+                )
+
+        # Batch file not supported
+        if batch_path:
+            parser.error(f'Technology {args.x} does not support a batch file.')
+
+        if args.fragment_l is not None or args.fragment_s is not None:
+            parser.error(
+                '`--fragment-l` and `--fragment-s` may only be provided with '
+                '`BULK` and `SMARTSEQ2` technologies.'
+            )
+
     if args.workflow in {'lamanno', 'nucleus'}:
         # Smartseq can not be used with lamanno or nucleus.
-        if args.x.upper() == 'SMARTSEQ':
+        if args.x.upper() in ('SMARTSEQ', 'SMARTSEQ2', 'BULK', 'SMARTSEQ3'):
             parser.error(
                 f'Technology `{args.x}` can not be used with workflow {args.workflow}.'
             )
@@ -216,11 +483,12 @@ def parse_count(parser, args, temp_dir='tmp'):
             args.c2,
             args.x,
             args.o,
-            args.fastqs,
+            batch_path or args.fastqs,
             args.w,
             tcc=args.tcc,
             mm=args.mm,
             filter=args.filter,
+            filter_threshold=args.filter_threshold,
             threads=args.t,
             memory=args.m,
             overwrite=args.overwrite,
@@ -230,7 +498,11 @@ def parse_count(parser, args, temp_dir='tmp'):
             report=args.report,
             inspect=not args.no_inspect,
             nucleus=args.workflow == 'nucleus',
-            temp_dir=temp_dir
+            temp_dir=temp_dir,
+            strand=args.strand,
+            umi_gene=args.umi_gene,
+            em=args.em,
+            by_name=args.gene_names
         )
     else:
         if args.workflow == 'kite:10xFB' and args.x.upper() != '10XV3':
@@ -240,6 +512,10 @@ def parse_count(parser, args, temp_dir='tmp'):
 
         # Smart-seq
         if args.x.upper() == 'SMARTSEQ':
+            logger.warning(
+                f'Technology `{args.x}` will be deprecated in the next release. '
+                'Please read the release notes on GitHub for more information. '
+            )
             if args.dry_run:
                 parser.error(f'Technology `{args.x}` does not support dry run.')
 
@@ -257,16 +533,6 @@ def parse_count(parser, args, temp_dir='tmp'):
                         f'Argument `{arg}` is not supported for technology `{args.x}`. This argument will be ignored.'
                     )
 
-            # Check if batch TSV was provided.
-            batch_path = None
-            if len(args.fastqs) == 1:
-                try:
-                    with open_as_text(args.fastqs[0], 'r') as f:
-                        if not f.readline().startswith('@'):
-                            batch_path = args.fastqs[0]
-                except Exception:
-                    pass
-
             cells = {}
             if batch_path:
                 with open(batch_path, 'r') as f:
@@ -292,6 +558,9 @@ def parse_count(parser, args, temp_dir='tmp'):
                         fastqs[i + 1] if i + 1 < len(fastqs) else ''
                     )
                     cells[cell_id] = (fastq_1, fastq_2)
+            if not cells:
+                parser.error('No FASTQs found.')
+
             logger.info('Found the following FASTQs:')
             fastq_pairs = []
             cell_ids = []
@@ -317,17 +586,36 @@ def parse_count(parser, args, temp_dir='tmp'):
                 h5ad=args.h5ad,
                 temp_dir=temp_dir
             )
+        elif args.x.upper() == 'SMARTSEQ3':
+            count_smartseq3(
+                args.i,
+                args.g,
+                args.o,
+                args.fastqs,
+                tcc=args.tcc,
+                mm=args.mm,
+                temp_dir=temp_dir,
+                threads=args.t,
+                memory=args.m,
+                overwrite=args.overwrite,
+                loom=args.loom,
+                h5ad=args.h5ad,
+                inspect=not args.no_inspect,
+                strand=args.strand,
+                by_name=args.gene_names
+            )
         else:
             count(
                 args.i,
                 args.g,
                 args.x,
                 args.o,
-                args.fastqs,
+                batch_path or args.fastqs,
                 args.w,
                 tcc=args.tcc,
                 mm=args.mm,
                 filter=args.filter,
+                filter_threshold=args.filter_threshold,
                 kite='kite' in args.workflow,
                 FB='10xFB' in args.workflow,
                 threads=args.t,
@@ -338,26 +626,36 @@ def parse_count(parser, args, temp_dir='tmp'):
                 cellranger=args.cellranger,
                 report=args.report,
                 inspect=not args.no_inspect,
-                temp_dir=temp_dir
+                temp_dir=temp_dir,
+                fragment_l=args.fragment_l,
+                fragment_s=args.fragment_s,
+                paired=args.parity == 'paired',
+                strand=args.strand,
+                umi_gene=args.umi_gene,
+                em=args.em,
+                by_name=args.gene_names
             )
 
 
 COMMAND_TO_FUNCTION = {
+    'compile': parse_compile,
     'ref': parse_ref,
     'count': parse_count,
 }
 
 
-def setup_info_args(parser, parent):
+def setup_info_args(
+    parser: argparse.ArgumentParser, parent: argparse.ArgumentParser
+) -> argparse.ArgumentParser:
     """Helper function to set up a subparser for the `info` command.
 
-    :param parser: argparse parser to add the `info` command to
-    :type args: argparse.ArgumentParser
-    :param parent: argparse parser parent of the newly added subcommand.
-                   used to inherit shared commands/flags
-    :type args: argparse.ArgumentParser
-    :return: the newly added parser
-    :rtype: argparse.ArgumentParser
+    Args:
+        parser: Parser to add the `info` command to
+        parent: Parser parent of the newly added subcommand.
+            used to inherit shared commands/flags
+
+    Returns:
+        The newly added parser
     """
     parser_info = parser.add_parser(
         'info',
@@ -369,17 +667,121 @@ def setup_info_args(parser, parent):
     return parser_info
 
 
-def setup_ref_args(parser, parent):
+def setup_compile_args(
+    parser: argparse.ArgumentParser, parent: argparse.ArgumentParser
+) -> argparse.ArgumentParser:
+    """Helper function to set up a subparser for the `compile` command.
+
+    Args:
+        parser: Parser to add the `compile` command to
+        parent: Parser parent of the newly added subcommand.
+            used to inherit shared commands/flags
+
+    Returns:
+        The newly added parser
+    """
+    parser_compile = parser.add_parser(
+        'compile',
+        description='Compile `kallisto` and `bustools` binaries from source',
+        help='Compile `kallisto` and `bustools` binaries from source',
+        parents=[parent],
+        add_help=False,
+    )
+
+    parser_compile.add_argument(
+        'target',
+        metavar='target',
+        help=(
+            'Which binaries to compile. May be one of `kallisto`, `bustools` or '
+            '`all`.'
+        ),
+        choices=['kallisto', 'bustools', 'all'],
+        default=None,
+        nargs='?',
+    )
+    compile_group = parser_compile.add_mutually_exclusive_group()
+    compile_group.add_argument(
+        '--view',
+        help=(
+            'See information about the current binaries, which are what will be '
+            'used for `ref` and `count`.'
+        ),
+        action='store_true',
+    )
+    compile_group.add_argument(
+        '--remove',
+        help=(
+            'Remove the existing compiled binaries. Binaries that are provided '
+            'with kb are never removed.'
+        ),
+        action='store_true',
+    )
+    compile_group.add_argument(
+        '--overwrite',
+        help='Overwrite the existing compiled binaries, if they exist.',
+        action='store_true',
+    )
+
+    parser_compile.add_argument(
+        '-o',
+        metavar='OUT',
+        help=(
+            'Save the compiled binaries to a different directory. Note that if this '
+            'option is specified, the binaries will have to be manually specified '
+            'with `--kallisto` or `--bustools` when running `ref` or `count`.'
+        ),
+        type=str,
+        default=None,
+    )
+    compile_group = parser_compile.add_mutually_exclusive_group()
+    compile_group.add_argument(
+        '--url',
+        metavar='URL',
+        help=(
+            'Use a custom URL to a ZIP or tarball file containing the source code '
+            'of the specified binary. May only be used with a single `target`.'
+        ),
+        type=str
+    )
+    compile_group.add_argument(
+        '--ref',
+        metavar='REF',
+        help=(
+            'Repository commmit hash or tag to fetch the source code from. '
+            'May only be used with a single `target`.'
+        ),
+        type=str
+    )
+    parser_compile.add_argument(
+        '--cmake-arguments',
+        metavar='URL',
+        help=(
+            'Additional arguments to pass to the cmake command. For example, to '
+            'pass additional include directories, '
+            '`--cmake-arguments="-DCMAKE_CXX_FLAGS=\'-I /usr/include\'"`'
+        ),
+        type=str,
+    )
+
+    return parser_compile
+
+
+def setup_ref_args(
+    parser: argparse.ArgumentParser, parent: argparse.ArgumentParser
+) -> argparse.ArgumentParser:
     """Helper function to set up a subparser for the `ref` command.
 
-    :param parser: argparse parser to add the `ref` command to
-    :type args: argparse.ArgumentParser
-    :param parent: argparse parser parent of the newly added subcommand.
-                   used to inherit shared commands/flags
-    :type args: argparse.ArgumentParser
-    :return: the newly added parser
-    :rtype: argparse.ArgumentParser
+    Args:
+        parser: Parser to add the `ref` command to
+        parent: Parser parent of the newly added subcommand.
+            used to inherit shared commands/flags
+
+    Returns:
+        The newly added parser
     """
+    kallisto_path = get_kallisto_binary_path()
+    bustools_path = get_bustools_binary_path()
+
     workflow = 'standard'
     for i, arg in enumerate(sys.argv):
         if arg.startswith('--workflow'):
@@ -425,6 +827,28 @@ def setup_ref_args(parser, parent):
         type=str,
         required='-d' not in sys.argv
     )
+    filter_group = parser_ref.add_mutually_exclusive_group()
+    filter_group.add_argument(
+        '--include-attribute',
+        metavar='KEY:VALUE',
+        help=(
+            'Only process GTF entries that have the provided KEY:VALUE attribute. '
+            'May be specified multiple times.'
+        ),
+        type=str,
+        action='append',
+    )
+    filter_group.add_argument(
+        '--exclude-attribute',
+        metavar='KEY:VALUE',
+        help=(
+            'Only process GTF entires that do not have the provided KEY:VALUE attribute. '
+            'May be specified multiple times.'
+        ),
+        type=str,
+        action='append',
+    )
+
     required_lamanno = parser_ref.add_argument_group(
         'required arguments for `lamanno` and `nucleus` workflows'
     )
@@ -504,6 +928,18 @@ def setup_ref_args(parser, parent):
         help='Overwrite existing kallisto index',
         action='store_true'
     )
+    parser_ref.add_argument(
+        '--kallisto',
+        help=f'Path to kallisto binary to use (default: {kallisto_path})',
+        type=str,
+        default=kallisto_path
+    )
+    parser_ref.add_argument(
+        '--bustools',
+        help=f'Path to bustools binary to use (default: {bustools_path})',
+        type=str,
+        default=bustools_path
+    )
     parser_ref.add_argument(
         'fasta',
         help='Genomic FASTA file(s), comma-delimited',
@@ -534,17 +970,22 @@ def setup_ref_args(parser, parent):
     return parser_ref
 
 
-def setup_count_args(parser, parent):
+def setup_count_args(
+    parser: argparse.ArgumentParser, parent: argparse.ArgumentParser
+) -> argparse.ArgumentParser:
     """Helper function to set up a subparser for the `count` command.
 
-    :param parser: argparse parser to add the `count` command to
-    :type args: argparse.ArgumentParser
-    :param parent: argparse parser parent of the newly added subcommand.
-                   used to inherit shared commands/flags
-    :type args: argparse.ArgumentParser
-    :return: the newly added parser
-    :rtype: argparse.ArgumentParser
+    Args:
+        parser: Parser to add the `count` command to
+        parent: Parser parent of the newly added subcommand.
+            used to inherit shared commands/flags
+
+    Returns:
+        The newly added parser
     """
+    kallisto_path = get_kallisto_binary_path()
+    bustools_path = get_bustools_binary_path()
+
     workflow = 'standard'
     for i, arg in enumerate(sys.argv):
         if arg.startswith('--workflow'):
@@ -619,6 +1060,13 @@ def setup_count_args(parser, parent):
         type=str,
         default='4G'
     )
+    parser_count.add_argument(
+        '--strand',
+        help='Strandedness (default: see `kb --list`)',
+        type=str,
+        default=None,
+        choices=['unstranded', 'forward', 'reverse']
+    )
     parser_count.add_argument(
         '--workflow',
         help=(
@@ -633,11 +1081,24 @@ def setup_count_args(parser, parent):
         default='standard',
         choices=['standard', 'lamanno', 'nucleus', 'kite', 'kite:10xFB']
     )
+    parser_count.add_argument(
+        '--em',
+        help='Estimate gene abundances using an EM algorithm.',
+        action='store_true'
+    )
+    parser_count.add_argument(
+        '--umi-gene',
+        help='Perform gene-level collapsing of UMIs.',
+        action='store_true'
+    )
 
     count_group = parser_count.add_mutually_exclusive_group()
     count_group.add_argument(
         '--mm',
-        help='Include reads that pseudoalign to multiple genes.',
+        help=(
+            'Include reads that pseudoalign to multiple genes. '
+            'Automatically enabled when generating a TCC matrix.'
+        ),
         action='store_true'
     )
     count_group.add_argument(
@@ -653,6 +1114,13 @@ def setup_count_args(parser, parent):
         nargs='?',
         choices=['bustools']
     )
+    parser_count.add_argument(
+        '--filter-threshold',
+        metavar='THRESH',
+        help='Barcode filter threshold (default: auto)',
+        type=int,
+        default=None,
+    )
     required_lamanno = parser_count.add_argument_group(
         'required arguments for `lamanno` and `nucleus` workflows'
     )
@@ -662,7 +1130,6 @@ def setup_count_args(parser, parent):
         help='Path to cDNA transcripts-to-capture',
         type=str,
         required=workflow in {'lamanno', 'nucleus'}
-        or any(arg in sys.argv for arg in {'--lamanno', '--nucleus'})
     )
     required_lamanno.add_argument(
         '-c2',
@@ -670,7 +1137,6 @@ def setup_count_args(parser, parent):
         help='Path to intron transcripts-to-captured',
         type=str,
         required=workflow in {'lamanno', 'nucleus'}
-        or any(arg in sys.argv for arg in {'--lamanno', '--nucleus'})
     )
     parser_count.add_argument(
         '--overwrite',
@@ -695,6 +1161,15 @@ def setup_count_args(parser, parent):
         help='Convert count matrices to cellranger-compatible format',
         action='store_true'
     )
+    parser_count.add_argument(
+        '--gene-names',
+        help=(
+            'Group counts by gene names instead of gene IDs when generating '
+            'the loom or h5ad file'
+        ),
+        action='store_true'
+    )
+
     report_group = parser_count.add_mutually_exclusive_group()
     report_group.add_argument(
         '--report',
@@ -708,9 +1183,50 @@ def setup_count_args(parser, parent):
     report_group.add_argument(
         '--no-inspect', help=argparse.SUPPRESS, action='store_true'
     )
+    parser_count.add_argument(
+        '--kallisto',
+        help=f'Path to kallisto binary to use (default: {kallisto_path})',
+        type=str,
+        default=kallisto_path
+    )
+    parser_count.add_argument(
+        '--bustools',
+        help=f'Path to bustools binary to use (default: {bustools_path})',
+        type=str,
+        default=bustools_path
+    )
     parser_count.add_argument(
         '--no-validate', help=argparse.SUPPRESS, action='store_true'
     )
+
+    optional_bulk = parser_count.add_argument_group(
+        'optional arguments for `BULK` and `SMARTSEQ2` technologies'
+    )
+    optional_bulk.add_argument(
+        '--parity',
+        help=(
+            'Parity of the input files. Choices are `single` for single-end '
+            'and `paired` for paired-end reads.'
+        ),
+        type=str,
+        choices=['single', 'paired'],
+        default=None
+    )
+    optional_bulk.add_argument(
+        '--fragment-l',
+        metavar='L',
+        help='Mean length of fragments. Only for single-end.',
+        type=int,
+        default=None
+    )
+    optional_bulk.add_argument(
+        '--fragment-s',
+        metavar='S',
+        help='Standard deviation of fragment lengths. Only for single-end.',
+        type=int,
+        default=None
+    )
+
     parser_count.add_argument(
         'fastqs',
         help=(
@@ -729,10 +1245,6 @@ def setup_count_args(parser, parent):
 def main():
     """Command-line entrypoint.
     """
-    # Get prepackaged kallisto and bustools paths.
-    kallisto_path = get_kallisto_binary_path()
-    bustools_path = get_bustools_binary_path()
-
     # Main parser
     parser = argparse.ArgumentParser(
         description='kb_python {}'.format(__version__)
@@ -764,25 +1276,15 @@ def main():
     parent.add_argument(
         '--verbose', help='Print debugging information', action='store_true'
     )
-    parent.add_argument(
-        '--kallisto',
-        help=f'Path to kallisto binary to use (default: {kallisto_path})',
-        type=str,
-        default=kallisto_path
-    )
-    parent.add_argument(
-        '--bustools',
-        help=f'Path to bustools binary to use (default: {bustools_path})',
-        type=str,
-        default=bustools_path
-    )
 
     # Command parsers
     setup_info_args(subparsers, argparse.ArgumentParser(add_help=False))
+    parser_compile = setup_compile_args(subparsers, parent)
     parser_ref = setup_ref_args(subparsers, parent)
     parser_count = setup_count_args(subparsers, parent)
 
     command_to_parser = {
+        'compile': parser_compile,
         'ref': parser_ref,
         'count': parser_count,
     }
@@ -828,14 +1330,37 @@ def main():
     logger.debug('Printing verbose output')
 
     # Set binary paths
-    set_kallisto_binary_path(args.kallisto)
-    set_bustools_binary_path(args.bustools)
+    if args.command in ('ref', 'count') and ('dry_run' not in args
+                                             or not args.dry_run):
+        if args.kallisto:
+            set_kallisto_binary_path(args.kallisto)
+        if args.bustools:
+            set_bustools_binary_path(args.bustools)
 
-    logger.debug(f'kallisto binary located at {get_kallisto_binary_path()}')
-    logger.debug(f'bustools binary located at {get_bustools_binary_path()}')
+        # Check
+        kallisto_path = get_kallisto_binary_path()
+        bustools_path = get_bustools_binary_path()
+        kallisto_ok, bustools_ok = test_binaries()
+
+        if not kallisto_path or not kallisto_ok:
+            raise UnsupportedOSError(
+                'Failed to find compatible kallisto binary. '
+                'Provide a compatible binary with the `--kallisto` option or '
+                'run `kb compile`.'
+            )
+        if not bustools_path or not bustools_ok:
+            raise UnsupportedOSError(
+                'Failed to find compatible bustools binary. '
+                'Provide a compatible binary with the `--bustools` option or '
+                'run `kb compile`.'
+            )
+
+        logger.debug(f'kallisto binary located at {kallisto_path}')
+        logger.debug(f'bustools binary located at {bustools_path}')
 
     temp_dir = args.tmp or (
-        os.path.join(args.o, TEMP_DIR) if 'o' in args else TEMP_DIR
+        os.path.join(args.o, TEMP_DIR)
+        if 'o' in args and args.o is not None else TEMP_DIR
     )
     # Check if temp_dir exists and exit if it does.
     # This is so that kb doesn't accidently use an existing directory and
